#version 450
 
layout (binding = 0) uniform UniformBuffer {
	vec4 color;
	mat4 modelMatrix;
	mat4 modelMatrixForNormals;
	mat4 viewProjMatrix;
	vec4 cameraPosition;
	vec4 illumination; // ka, kd, ks, alpha
	float brighntess;
	float time;
} ub_data;

layout (binding = 1) uniform DirectionalLight {
	vec4 color;
	vec4 direction;
	mat4 depthViewProjMatrix;
} dl_data;

layout (binding = 2) uniform PointLight {
	vec4 color;
	vec4 position;
	vec4 attenuation;
} pl_data;

layout (binding = 3) uniform sampler2D albedo_texture;
layout (binding = 4) uniform sampler2D normal_texture;
layout (binding = 5) uniform sampler2D metallic_texture;
layout (binding = 6) uniform sampler2D roughness_texture;
layout (binding = 7) uniform sampler2D ao_texture;
layout (binding = 8) uniform sampler2D emit_texture;
layout (binding = 9) uniform samplerCube irradiance_cubemap;
layout (binding = 10) uniform samplerCube prefilterMap;
layout (binding = 11) uniform sampler2D   brdfLUT;

// Shadowmapping
layout (binding = 12) uniform sampler2D shadowMap;


layout (location = 0) in VertexData {
	vec3 positionWorld;
	vec3 normalWorld;
	vec2 textureCoordinates;
	// Shadowmapping
	vec4 shadowCoordinates;
} frag_in;

layout (location = 0) out vec4 out_color;


// ---------------------------- Shadowmapping ---------------------------- //
#define shadow_ambient 0.01f
#define enablePCF 1
#define shadow_bias_offset_min 0.005f
#define shadow_bias_offset_max 0.05f

float textureProj(vec4 shadowCoord, vec2 off)
{
	float shadow = 1.0;
	if ( shadowCoord.z > -1.0 && shadowCoord.z < 1.0 )
	{
		float dist = texture( shadowMap, shadowCoord.st + off ).r;
		// shadow bias to prevent shadow acne
		float shadow_bias = max(shadow_bias_offset_max * (1.0 - dot(frag_in.normalWorld, -dl_data.direction.xyz)), shadow_bias_offset_min);

		if ( shadowCoord.w > 0.0 && dist + shadow_bias < shadowCoord.z )
		{
			shadow = shadow_ambient;
		}
	}
	return shadow;
}

float filterPCF(vec4 sc)
{
	ivec2 texDim = textureSize(shadowMap, 0);
	float scale = 1.5;
	float dx = scale * 1.0 / float(texDim.x);
	float dy = scale * 1.0 / float(texDim.y);

	float shadowFactor = 0.0;
	int count = 0;
	int range = 1;

	for (int x = -range; x <= range; x++)
	{
		for (int y = -range; y <= range; y++)
		{
			shadowFactor += textureProj(sc, vec2(dx*x, dy*y));
			count++;
		}

	}
	return shadowFactor / count;
}


// ---------------------------- PBR ---------------------------- //

const float PI = 3.14159265359;

vec3 getNormalFromNormalMap() {
	vec3 tangentNormal = texture(normal_texture, frag_in.textureCoordinates).xyz * 2.0 - 1.0;

	vec3 Q1  = dFdx(frag_in.positionWorld);
	vec3 Q2  = dFdy(frag_in.positionWorld);
	vec2 st1 = dFdx(frag_in.textureCoordinates);
	vec2 st2 = dFdy(frag_in.textureCoordinates);

	vec3 N   = normalize(frag_in.normalWorld);
	vec3 T  = normalize(Q1*st2.t - Q2*st1.t);
	vec3 B  = -normalize(cross(N, T));
	mat3 TBN = mat3(T, B, N);

	return normalize(TBN * tangentNormal);
}

float DistributionGGX(vec3 N, vec3 H, float roughness) {
	float a = roughness*roughness;
	float a2 = a*a;
	float NdotH = max(dot(N, H), 0.0);
	float NdotH2 = NdotH*NdotH;

	float nom   = a2;
	float denom = (NdotH2 * (a2 - 1.0) + 1.0);
	denom = PI * denom * denom;

	return nom / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness) {
	float r = (roughness + 1.0);
	float k = (r*r) / 8.0;

	float nom   = NdotV;
	float denom = NdotV * (1.0 - k) + k;

	return nom / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
	float NdotV = max(dot(N, V), 0.0);
	float NdotL = max(dot(N, L), 0.0);
	float ggx2 = GeometrySchlickGGX(NdotV, roughness);
	float ggx1 = GeometrySchlickGGX(NdotL, roughness);

	return ggx1 * ggx2;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
	return F0 + (1.0 - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}

vec3 fresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness)
{
	return F0 + (max(vec3(1.0 - roughness), F0) - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}


vec3 Li(vec3 V, vec3 N, vec3 L, vec3 albedo, float metallic, float roughness, vec3 F0, vec3 lightColor, bool attenuate, vec3 attenuation) {
	float lightDist = length(L);
	L = normalize(L);
	vec3 H = normalize(V + L);
	
	float recip_attenuation = 1.0;
	if(attenuate) {
		recip_attenuation = 1.0f / (attenuation.x + lightDist * attenuation.y + lightDist * lightDist * attenuation.z);
	}
	vec3 radiance = lightColor * recip_attenuation;

	// Cook-Torrance BRDF
	float NDF = DistributionGGX(N, H, roughness);
	float G = GeometrySmith(N, V, L, roughness);
	vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);

	vec3 numerator    = NDF * G * F;
	float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001; // + 0.0001 to prevent divide by zero
	vec3 specular = numerator / denominator;

	// kS is equal to Fresnel
	vec3 kS = F;
	// for energy conservation, the diffuse and specular light can't
	// be above 1.0 (unless the surface emits light); to preserve this
	// relationship the diffuse component (kD) should equal 1.0 - kS.
	vec3 kD = vec3(1.0) - kS;
	// multiply kD by the inverse metalness such that only non-metals 
	// have diffuse lighting, or a linear blend if partly metal (pure metals
	// have no diffuse light).
	kD *= 1.0 - metallic;

	// scale light by NdotL
	float NdotL = max(dot(N, L), 0.0);

	// add to outgoing radiance Lo
	return (kD * albedo / PI + specular) * radiance * NdotL;  // note that we already multiplied the BRDF by the Fresnel (kS) so we won't multiply by kS again
}


vec3 prefilteredReflection(vec3 r, float roughness)
{
	const float MAX_REFLECTION_LOD = 9.0;
	float lod = roughness * MAX_REFLECTION_LOD;
	float lodf = floor(lod);
	float lodc = ceil(lod);
	vec3 a = textureLod(prefilterMap, r, lodf).rgb;
	vec3 b = textureLod(prefilterMap, r, lodc).rgb;
	return mix(a, b, lod - lodf);
}


void main() {
	vec3 n = getNormalFromNormalMap();
	vec3 v = normalize(ub_data.cameraPosition.xyz - frag_in.positionWorld);
	
	vec3 albedo = pow(texture(albedo_texture, frag_in.textureCoordinates).rgb, vec3(2.2));
	float metallic  = texture(metallic_texture, frag_in.textureCoordinates).r;
	float roughness = texture(roughness_texture, frag_in.textureCoordinates).r;
	float ao        = texture(ao_texture, frag_in.textureCoordinates).r;

	vec3 F0 = vec3(0.04);
	F0 = mix(F0, albedo, metallic);
	
	vec3 Lo = vec3(0.0);
	
	// Add directional light's contribution:
	Lo += Li(v, n, -dl_data.direction.xyz, albedo, metallic, roughness, F0, dl_data.color.rgb, false, vec3(1.0));

	// Add point light's contribution
	Lo += Li(v, n, pl_data.position.xyz - frag_in.positionWorld, albedo, metallic, roughness, F0, pl_data.color.rgb, true, pl_data.attenuation.xyz);

	// Ambient light
	//legacy code: vec3 ambient = vec3(0.03) * albedo * ao;

	// Diffuse irradiance contribution to ambient light
	vec3 kS = fresnelSchlickRoughness(max(dot(n, v), 0.0), F0, roughness);
	vec3 kD = (1.0 - kS) * (1.0 - metallic);

	vec3 irradiance = texture(irradiance_cubemap, n).rgb;
	vec3 diffuse    = irradiance * albedo;

	// Specular reflection via prefiltered cubemap and BRDF 2D LUT
	vec3 r = reflect(-v, n);
	vec3 prefilteredColor  = prefilteredReflection(r, roughness);

	vec2 envBRDF  = texture(brdfLUT, vec2(max(dot(n, v), 0.0), roughness)).rg;
	vec3 specular = prefilteredColor * (kS * envBRDF.x + envBRDF.y);
	vec3 ambient  = (kD * diffuse + specular) * ao;
	
	vec3 color = ambient + Lo;

	//Shadowmapping
	float shadow = (enablePCF == 1) ? filterPCF(frag_in.shadowCoordinates / frag_in.shadowCoordinates.w) : textureProj(frag_in.shadowCoordinates / frag_in.shadowCoordinates.w, vec2(0.0));
	color *= shadow;

	// HDR tonemapping
	color = color / (color + vec3(1.0));
	// gamma correct
	color = pow(color, vec3(1.0/2.2));

	
	out_color = vec4(color, 1.0);
}

